/* See LICENSE file for license details. */

#include <errno.h>
#include <unistd.h>
#include <security/pam_appl.h>
#include <X11/Xft/Xft.h>

#include "../config.h"
#include "../slock.h"

struct AuthState {
	XftColor color_error;
	XftColor color_info;
	XftFont *font;
	struct MsgList *msglist;
	short echo_offset;
};

struct MsgList {
	char *msg;
	int style;
	const struct MsgList *prev;
};

struct AuthList {
	XftDraw *draw;
};

static int
get_window_height(const struct State *const state)
{
	XWindowAttributes wa;

	XGetWindowAttributes(state->dpy, state->wio->win, &wa);
	return wa.height;
}

static struct MsgList *const
append_msg(struct MsgList *const msglist, char *const msg, const int style)
{
	struct MsgList *m;

	m = malloc(sizeof(struct MsgList));
	if (!m) {
		eprintln(strerror(errno));
		return msglist;
	}

	m->msg = msg;
	m->style = style;
	m->prev = msglist;

	return m;
}

static int
pam_conv(int num_msg, const struct pam_message **const messages,
		struct pam_response **const responses,
		void *const appdata_ptr)
{
	XGlyphInfo extents;
	char *msg;
	int i;

	struct State *const state = appdata_ptr;

	for (i = 0; i < num_msg; i++) {
		msg = malloc(strlen(messages[i]->msg) + (messages[i]->msg_style
			== PAM_PROMPT_ECHO_ON ? sizeof(state->input) : 1));
		if (!msg) {
			eprintln(strerror(errno));
			return PAM_BUF_ERR;
		}
		strcpy(msg, messages[i]->msg);
		state->auth->msglist = append_msg(state->auth->msglist, msg,
			messages[i]->msg_style);
		redraw(state);

		responses[i] = malloc(sizeof(struct pam_response));
		if (!responses[i]) {
			eprintln(strerror(errno));
			return PAM_BUF_ERR;
		}
		responses[i]->resp_retcode = 0;
		responses[i]->resp = malloc(sizeof(state->input));
		if (!responses[i]->resp) {
			eprintln(strerror(errno));
			return PAM_BUF_ERR;
		}

		if (state->auth->msglist->style != PAM_PROMPT_ECHO_OFF &&
				state->auth->msglist->style !=
				PAM_PROMPT_ECHO_ON)
			continue;

		if (state->auth->msglist->style == PAM_PROMPT_ECHO_ON) {
			XftTextExtentsUtf8(state->dpy, state->auth->font,
				(const unsigned char *)
					state->auth->msglist->msg,
				strlen(state->auth->msglist->msg), &extents);
			state->inputend = NULL;
			explicit_bzero(state->input, sizeof(state->input));
			state->intent = Read;
			state->auth->echo_offset = extents.xOff;
			get_events(state);
			state->auth->echo_offset = -1;
			if (state->intent == Submit)
				strcat(state->auth->msglist->msg,
					state->input);
		} else if (!*state->input) {
			state->intent = Read;
			get_events(state);
		}
		if (state->intent == Reset)
			return PAM_CONV_ERR;
		strcpy(responses[i]->resp, state->input);
		state->inputend = NULL;
		explicit_bzero(state->input, sizeof(state->input));
	}

	return PAM_SUCCESS;
}

void
auth_redraw(const struct State *const state)
{
	const struct MsgList *m;
	int i;

	for (m = state->auth->msglist, i = 1; m; m = m->prev, i++)
		XftDrawStringUtf8(
			state->wio->auth->draw,
			m->style == PAM_ERROR_MSG ?
				&state->auth->color_error :
				&state->auth->color_info,
			state->auth->font,
			state->auth->font->height,
			get_window_height(state) - i *
				state->auth->font->height,
			(const unsigned char *) m->msg,
			strlen(m->msg));
}

void
auth_reset(const struct State *const state, char *const msg, const int fail)
{
	if (state->auth->msglist || fail)
		state->auth->msglist = append_msg(state->auth->msglist,
			msg, fail ? PAM_ERROR_MSG : PAM_TEXT_INFO);
}

void
auth_backspace(const struct State *const state)
{
	XGlyphInfo extents;

	/* Remove echoed input. */
	if (state->auth->echo_offset >= 0) {
		XftTextExtentsUtf8(
			state->dpy,
			state->auth->font,
			(const unsigned char *) state->input,
			state->inputend - state->input,
			&extents);
		XClearArea(
			state->dpy,
			state->wio->win,
			state->auth->font->height + state->auth->echo_offset +
				extents.xOff,
			get_window_height(state) - state->auth->font->height -
				state->auth->font->ascent,
			state->auth->font->max_advance_width,
			state->auth->font->ascent + state->auth->font->descent,
			False);
	}
}

void
auth_input(const struct State *const state, const char *const buf,
		const int len)
{
	XGlyphInfo extents;

	/* Echo the input if requested. */
	if (state->auth->echo_offset >= 0) {
		XftTextExtentsUtf8(
			state->dpy,
			state->auth->font,
			(const unsigned char *) state->input,
			state->inputend - state->input,
			&extents);
		XftDrawStringUtf8(
			state->wio->auth->draw,
			&state->auth->color_info,
			state->auth->font,
			state->auth->font->height + state->auth->echo_offset +
				extents.xOff,
			get_window_height(state) -
				state->auth->font->height,
			(const unsigned char *) buf,
			len);
	}
}

void
auth_init(struct State *const state)
{
	XWindowAttributes wa;

	state->auth = malloc(sizeof(*state->auth));
	state->auth->msglist = NULL;
	state->auth->echo_offset = -1;

	/* color_error */
	state->auth->color_error.pixel = COLOR_ERROR;
	state->auth->color_error.color = (XRenderColor) {
		COLOR_ERROR >> 8 & 0xff00,
		COLOR_ERROR      & 0xff00,
		COLOR_ERROR << 8 & 0xff00,
		0xffff};

	/* color_info */
	state->auth->color_info.pixel = COLOR_INFO;
	state->auth->color_info.color = (XRenderColor) {
		COLOR_INFO >> 8 & 0xff00,
		COLOR_INFO      & 0xff00,
		COLOR_INFO << 8 & 0xff00,
		0xffff};

	/* font */
	state->auth->font = XftFontOpenName(state->dpy, 0,
		"monospace:size=10");
	check_val(state->auth->font != NULL, "XftFontOpenName",
		"unable to load font");

	for (state->wio = state->wios; state->wio;
			state->wio = state->wio->next) {
		XGetWindowAttributes(state->dpy, state->wio->win, &wa);
		state->wio->auth = malloc(sizeof(*state->wio->auth));
		state->wio->auth->draw = XftDrawCreate(state->dpy,
			state->wio->win, wa.visual, wa.colormap);
	}
}

int
auth_do(struct State *const state)
{
	int rv;
	pam_handle_t *pamh;
	const struct pam_conv pamc = {pam_conv, state};

	rv = pam_start(PAM_SERVICE_NAME, getlogin(), &pamc, &pamh);
	if (rv == PAM_SUCCESS)
		rv = pam_authenticate(pamh, 0);
	pam_end(pamh, rv);
	if (rv != PAM_SUCCESS && rv != PAM_AUTHTOK_ERR)
		reset(state, (char *) pam_strerror(pamh, rv), 1);
	return rv;
}

void
auth_close(const struct State *const state)
{
}
