/* See LICENSE file for license details. */

#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "config.h"
#include "slock.h"

static void
handle_keypress(XKeyEvent *const ev, struct State *const state)
{
	Status status;
	char buf[32];
	int rv;

	rv = XmbLookupString(state->wio->ic, ev, buf, sizeof(buf), NULL,
		&status);

	switch (status) {
	case XBufferOverflow:  /* FALLTHROUGH */
		reset(state, "XBufferOverflow", 1);
	case XLookupNone:  /* Meta Keys */
		return;
	}

	switch (buf[0]) {
	case 0x8:  /* Backspace */
		if (!state->input[0]) {
			state->intent = Reset;
			redraw(state);
			usleep(100000);
			state->intent = Read;
			redraw(state);
			break;
		}

		/* Remove last character from string, taking into account UTF-8
		 * continuation bytes. */
		state->inputend--;
		while ((*state->inputend & 0xc0) == 0x80)
			state->inputend--;
		*state->inputend = '\0';

		auth_backspace(state);

		break;
	case 0xd:  /* Return or Enter */
		state->intent = Submit;
		break;
	case 0x1b:  /* Escape */
		if (state->intent == Read)
			reset(state, "Aborted", 0);
		break;
	default:
		/* Ensure intent is set. */
		if (state->intent != Read) {
			state->intent = Read;
			redraw(state);
		}

		if (!state->inputend)
			state->inputend = state->input;

		/* Abort with error if control characters are input. */
		if (iscntrl(buf[0]) || state->inputend + rv >= state->input +
				sizeof(state->input)) {
			reset(state, "Illegal Input", 1);
			break;
		}

		auth_input(state, buf, rv);

		/* Store the new input. */
		memcpy(state->inputend, buf, rv);
		state->inputend += rv;
	}
}

static void
handle_visibilitynotify(const XVisibilityEvent *const ev)
{
	XWindowAttributes wa;

	/* Nothing happened that requires handling when the window becomes
	 * unobscured. */
	if (ev->state == VisibilityUnobscured)
		return;

	/* Any kind of window obscuration requires immediately raising the
	 * window. */
	XRaiseWindow(ev->display, ev->window);
	XSync(ev->display, False);

	/* That is all that needs to be done, unless the window was fully
	 * obscured. */
	if (ev->state != VisibilityFullyObscured)
		return;

	/* Resizing the display fully obscures the window for a moment.
	 * Therefore, we will ensure that its size matches the display's size
	 * for all such events. */
	XGetWindowAttributes(ev->display, ev->window, &wa);
	XGetWindowAttributes(ev->display, wa.root, &wa);
	XResizeWindow(ev->display, ev->window, wa.width, wa.height);
	XSync(ev->display, False);
}

static void
handle_default(const XAnyEvent *const ev)
{
	eprintln("unexepected event: type=%i, serial=%lu, send_event=%i, "
			"display=0x%lx, window=0x%lx",
		ev->type, ev->serial, ev->send_event, ev->display, ev->window);
}

static void
init(struct State *const state)
{
	Cursor cursor;
	Pixmap pmap;
	XColor color;
	XSetWindowAttributes swa;
	char *rvl;
	int i, j, n, rv;

	/* Trivial initial values in state */
	state->intent = Reset;
	state->reset_color = COLOR_INIT;
	state->inputend = NULL;
	state->wios = NULL;
	state->wio = NULL;

	/* dpy */
	rvl = setlocale(LC_ALL, "");
	check_val(rvl != NULL, "setlocale", "unable to set locale");
	rvl = XSetLocaleModifiers("");
	check_val(rvl != NULL, "XSetLocaleModifiers",
		"unable to set locale modifiers");
	state->dpy = XOpenDisplay(NULL);
	check_val(state->dpy != NULL, "XOpenDisplay", "unable to open display");

	/* input */
	explicit_bzero(state->input, sizeof(state->input));

	/* Initialize all screens. */
	swa.background_pixel = COLOR_INIT;
	swa.event_mask = VisibilityChangeMask | ExposureMask;
	swa.override_redirect = 1;
	n = ScreenCount(state->dpy);
	check_val(n, "ScreenCount", "unable to determine number of screens");
	for (i = 0; i < n; i++) {
		/* wio and wios */
		state->wio = *(state->wio ? &state->wio->next : &state->wios) =
			malloc(sizeof(struct WinIOList));
		check_val(state->wio != NULL, "malloc", strerror(errno));

		/* wio->next */
		state->wio->next = NULL;

		/* wio->win */
		state->wio->win = XCreateWindow(state->dpy,
			RootWindow(state->dpy, i), 0, 0,
			DisplayWidth(state->dpy, i),
			DisplayHeight(state->dpy, i), 0, CopyFromParent,
			CopyFromParent, CopyFromParent,
			CWBackPixel | CWEventMask | CWOverrideRedirect, &swa);
		check_val(state->wio->win, "XCreateWindow",
			"unable to create window");

		/* wio->ic */
		state->wio->ic = XCreateIC(
			XOpenIM(state->dpy, NULL, NULL, NULL), XNInputStyle,
			XIMPreeditNothing | XIMStatusNothing, XNClientWindow,
			state->wio->win, NULL);
		check_val(state->wio->ic != NULL, "XCreateIC",
			"unable to create input context");

		/* Make cursor invisible. */
		pmap = XCreateBitmapFromData(state->dpy, state->wio->win,
			"\0\0\0\0\0\0\0\0", 8, 8);
		check_val(pmap, "XCreateBitmapFromData",
			"unable to create cursor bitmap");
		cursor = XCreatePixmapCursor(state->dpy, pmap, pmap, &color,
			&color, 0, 0);
		check_val(cursor, "XCreatePixmapCursor",
			"unable to create invisible cursor");

		/* Make window visible. */
		rv = XMapRaised(state->dpy, state->wio->win);
		check_val(rv, "XMapRaised", "unable to map window");

		/* Grab input. */
		for (j = 0;; j++) {
			rv = XGrabKeyboard(state->dpy, state->wio->win, False,
				GrabModeAsync, GrabModeAsync, CurrentTime);
			if (rv == GrabSuccess)
				break;
			if (j >= 6)
				check_val(0, "init", "unable to grab keyboard");
			usleep(100000);
		}
		for (j = 0;; j++) {
			rv = XGrabPointer(state->dpy, state->wio->win, False,
				None, GrabModeAsync, GrabModeAsync, None,
				cursor, CurrentTime);
			if (rv == GrabSuccess)
				break;
			if (j >= 6)
				check_val(0, "init", "unable to grab pointer");
			usleep(100000);
		}
	}

	rv = XSync(state->dpy, False);
	check_val(rv, "XSync", "unable to synchronize with X");
}

void
eprintln(const char *const fmt, ...)
{
	va_list ap;

	fprintf(stderr, "slock: ");
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}

void
check_val(const int val, const char *const fname, const char *const message)
{
	if (val)
		return;

	eprintln("%s: %s", fname ? fname : "ERROR", message);
	exit(1);
}

void
redraw(const struct State *const state)
{
	XSetWindowBackground(state->dpy, state->wio->win,
		state->intent == Reset ? state->reset_color : COLOR_INPUT);
	XClearWindow(state->dpy, state->wio->win);
	auth_redraw(state);
	XSync(state->dpy, False);
}

void
reset(struct State *const state, char *const msg, const int fail)
{
	if (fail) {
		eprintln(msg);
		XBell(state->dpy, 100);
		state->reset_color = COLOR_FAILED;
	}

	if (state->inputend) {
		state->inputend = NULL;
		explicit_bzero(state->input, sizeof(state->input));
	}

	auth_reset(state, msg, fail);
	state->intent = Reset;
	redraw(state);
}

void
get_events(struct State *const state)
{
	XEvent ev;

	while (!XNextEvent(state->dpy, &ev)) {
		if (XFilterEvent(&ev, None))
			continue;

		/* Track window input output structure (wio) of last event. */
		if (state->wio->win != ev.xany.window && ev.xany.window) {
			for (state->wio = state->wios; state->wio;
					state->wio = state->wio->next)
				if (state->wio->win == ev.xany.window)
					break;
			if (!state->wio)
				state->wio = state->wios;
		}

		switch (ev.type) {
		case KeyPress:
			handle_keypress(&ev.xkey, state);
			if (state->intent != Read)
				return;
			break;
		case KeyRelease:
			/* XGrabKeyboard() causes KeyPress *and* KeyRelease
			 * events to be reported, even when they are not
			 * selected in XSetWindowAttributes wa. Since we are
			 * not interested in KeyRelease events, we just quit
			 * the switch without doing anything. */
			break;
		case VisibilityNotify:
			handle_visibilitynotify(&ev.xvisibility);
			break;
		case Expose:
			redraw(state);
			break;
		case MappingNotify:
			/* If the modifier, keyboard, or pointer mapping is
			 * changed by any client in X, this event is sent to
			 * all X clients. It cannot be suppressed. It is
			 * ignored here. */
			break;
		default:
			handle_default(&ev.xany);
		}
	}
}

int
main(const int argc, const char *const *const argv)
{
	int rv;
	struct State state;

	if (argc != 1) {
		fprintf(stderr, "Usage: slock\n");
		return 1;
	}

	init(&state);
	auth_init(&state);
	state.wio = state.wios;

	rv = -1;
	do {
		get_events(&state);
		if (state.intent == Reset)
			continue;
		rv = auth_do(&state);
	} while (rv);

	auth_close(&state);

	return 0;
}
